from typing import List, Tuple, Union, Optional
from dataclasses import dataclass, field
from functools import reduce
from pathlib import Path

from .types import ConfigKey, ConfigValue
from .utils import list_uniq


@dataclass(frozen=True)
class ConfigSource:
    uri: str
    index: str
    loader: "ConfigLoader" = field(repr=False, default=None, hash=False)

    @property
    def name(self) -> str:
        return Path(self.uri).stem

    def __repr__(self) -> str:
        return f"{self.name}:{self.index}"

    def __add__(
        self, other: "ConfigSource"
    ) -> Union["ConfigSource", "MergedConfigSource"]:
        if self == other:
            return self
        return MergedConfigSource(sources=list([self])) + other

    def find(self, key: ConfigKey) -> Optional[ConfigValue]:
        if self.loader is None:
            return None
        return self.loader.find(self.uri, key)


@dataclass(frozen=True)
class MergedConfigSource:
    sources: List[ConfigSource]

    def __add__(
        self, other: Union[ConfigSource, "MergedConfigSource"]
    ) -> "MergedConfigSource":
        if isinstance(other, ConfigSource):
            source = list([other])
        if isinstance(other, MergedConfigSource):
            source = other.sources
        return MergedConfigSource(sources=list_uniq(self.sources + source))

    def __str__(self) -> str:
        def create_breadcrumb(a: Tuple[str, str], s: ConfigSource) -> Tuple[str, str]:
            name, breadcrumb = a
            if (n := s.name) == name:
                return n, f"{breadcrumb}-{s.index}"
            else:
                return n, f"{breadcrumb}-{n}-{s.index}"

        return reduce(create_breadcrumb, self.sources, ("", ""))[1].lstrip("-")

    def find(self, key: ConfigKey) -> Optional[ConfigValue]:
        for loader, uri in reversed([(s.loader, s.uri) for s in self.sources]):
            v = loader.find(uri, key)
            if v is not None:
                return v
        return None



from typing import List, Optional, Any, Union
from collections import OrderedDict
from operator import add
import re

from .config import Config
from .loader import ConfigLoader
from .snippet import ConfigSnippet
from .types import ConfigKey, ConfigFile, WILDCARD, ConfigName, ConfigNameGetter
from .utils import list_flatten, dict_merge_with_wildcard


class ConfigSet(OrderedDict):
    @classmethod
    def load(cls, *args: ConfigFile, **kwargs: Any) -> "ConfigSet":
        loader = ConfigLoader(*args, **kwargs)
        loader.load()
        snippets = list_flatten(loader.values())

        named_snippets = OrderedDict()
        for snippet in snippets:
            named_snippets.setdefault(
                get_config_name(snippet, kwargs.get("name", "${group}-${name}")), [],
            ).append(snippet)

        named_configs = {
            name: Config.from_snippets(snippets, **kwargs)
            for name, snippets in named_snippets.items()
        }
        return cls(named_configs)

    def __add__(self, other: "ConfigSet") -> "ConfigSet":
        return dict_merge_with_wildcard(self, other, add)

    def config_names(self) -> List[ConfigName]:
        return self.keys()

    def get_config(
        self, key: ConfigName, default: Optional[Config] = None
    ) -> Optional[Config]:
        return self.get(key, default)


def get_config_name_by_fields(
    snippet: ConfigSnippet, from_fields: List[ConfigKey]
) -> ConfigName:
    ids = [
        str(n) for n in [snippet.find(field) for field in from_fields] if n is not None
    ]
    if ids:
        return "-".join(ids)
    else:
        return WILDCARD


def get_config_name_by_statement(snippet: ConfigSnippet, getter: str):
    reg = r"\${(\w+)}"
    matches = re.findall(reg, getter)
    vars = {v: snippet.find(v) for v in matches}
    if (
        len(matches) > 0
        and len(
            [
                exist
                for exist in [vars.get(m, None) for m in matches]
                if exist is not None
            ]
        )
        == 0
    ):
        return WILDCARD

    def sub(var):
        (key,) = var.groups()
        return str(vars.get(key, None))

    return re.sub(reg, sub, getter)


def get_config_name(
    snippet: ConfigSnippet, getter: Union[str, ConfigNameGetter]
) -> ConfigName:
    if callable(getter):
        return getter(snippet)
    if isinstance(getter, str):
        return get_config_name_by_statement(snippet, getter)
    if isinstance(getter, list):
        return get_config_name_by_fields(snippet, getter)
    return WILDCARD



from abc import ABC, abstractmethod
from collections import OrderedDict
from os import path
from typing import List, Optional, Any

import yaml

from .snippet import ConfigSnippet
from .source import ConfigSource
from .types import ConfigFile, ConfigDict, ConfigValue, ConfigKey


class BaseConfigLoader(OrderedDict, ABC):
    """
    Key: the path of loaded yaml
    Value: the config dict of yaml
    """

    def __init__(self, *files: ConfigFile, **kwargs: Any):
        self._source = files
        self._includes = []

    @abstractmethod
    def _parse_config(self, content) -> List[ConfigDict]:
        ...

    def find(self, file: ConfigFile, key: ConfigKey) -> Optional[ConfigValue]:
        for snippet in self.get(file):
            v = snippet.get(key, None)
            if v is not None:
                return v
        return None

    def load(self) -> None:
        [self._load(file) for file in self._source]

    def _read_file(self, file: ConfigFile) -> str:
        key = path.abspath(file)
        if key in self.keys():
            return None
        with open(file, "r") as fh:
            return fh.read()

    def _load(self, file: ConfigFile, source: Optional[ConfigSource] = None) -> None:
        file = path.abspath(file)
        _conf = self._parse_config(self._read_file(file))

        configs = [self._process_directives(file, c) for c in _conf]
        configs = [dc for dc in configs if dc]

        def gs(config, file, index, ps=None):
            source = ConfigSource(uri=file, index=index, loader=self)
            if ps is not None:
                source = ps + source
            return ConfigSnippet(config=config, source=source)

        snippets = [
            gs(config, file, index, source) for index, config in enumerate(configs)
        ]
        self.setdefault(file, snippets)

        while self._includes:
            inc, source = self._includes.pop()
            self._load(inc, source)

    def _process_directives(self, file: ConfigFile, config: ConfigDict) -> ConfigDict:
        workdir = path.dirname(file)
        includes = config.pop("@include", None)
        if includes is not None:
            self._includes.extend(
                [
                    (
                        path.abspath(path.join(workdir, f)),
                        ConfigSource(uri=file, index=0, loader=self),
                    )
                    for f in includes
                ]
            )

        return config


class YamlConfigLoader(BaseConfigLoader):
    def _parse_config(self, content) -> List[ConfigDict]:
        if content is None:
            return []
        return yaml.load_all(content, Loader=yaml.SafeLoader)


class ConfigLoader(object):
    def __new__(cls, *args: ConfigFile, **kwargs: Any):
        return YamlConfigLoader(*args, **kwargs)



from typing import TypeVar, List, Union, Mapping, Callable

ConfigFile = str
ConfigName = str

ConfigKey = str
ConfigValue = Union[List[Union[str, int]], "ConfigDict", str, int]
ConfigDict = Mapping[ConfigKey, ConfigValue]

ProfileKey = str
ProfileName = str

WILDCARD = "*"
PROFILE_NAME_KEY = "profile"
CONFIG_NAME_KEY = ["group", "name"]
ConfigNameGetter = Callable[["ConfigSnippet"], ConfigName]

KT = TypeVar("KT")
VT = TypeVar("VT")



from functools import reduce
from operator import add
from typing import Any

from .snippet import ConfigSnippet
from .config import Config
from .configset import ConfigSet
from .types import ConfigFile


def configchain(*files: ConfigFile, **kwargs: Any) -> ConfigSet:
    return reduce(add, [ConfigSet.load(f, **kwargs) for f in files])



from collections import OrderedDict
from functools import reduce
from operator import add
from typing import List, Optional, Callable

from .snippet import ConfigSnippet
from .types import WILDCARD, ProfileKey, ProfileName, PROFILE_NAME_KEY
from .utils import list_groupby, dict_merge_with_wildcard


class Config(OrderedDict):
    @classmethod
    def from_snippets(cls, snippets: List[ConfigSnippet], **kwargs) -> "Config":
        profile = kwargs.get("profile", PROFILE_NAME_KEY)
        profile_getter: Callable[["ConfigSnippet"], ProfileName] = lambda x: x.get(
            profile, WILDCARD
        )

        def groupby_profile_and_merge(
            snippets: List[ConfigSnippet],
        ) -> List[ConfigSnippet]:
            return [reduce(add, g) for g in list_groupby(snippets, profile_getter)]

        profile_snippets = groupby_profile_and_merge(snippets)
        config = OrderedDict({profile_getter(s): s for s in profile_snippets})

        wp = config.get(WILDCARD, None)
        if wp is not None:
            config.update({p: wp + c for p, c in config.items() if p != WILDCARD})

        return Config(config)

    def __add__(self, other: "Config") -> "Config":
        return dict_merge_with_wildcard(self, other, add)

    def profile_names(self) -> List[ProfileName]:
        return self.keys()

    def get_profile(
        self, key: ProfileKey, default: Optional[ConfigSnippet] = WILDCARD
    ) -> Optional[ConfigSnippet]:
        return self.get(key, self.get(default, None))



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="configchain",  # Replace with your own username
    version="0.3.4",
    author="Luo Tao",
    author_email="lotreal@gmail.com",
    description="hierarchical configuration",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lotreal/configchain",
    packages=setuptools.find_packages(),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.6",
)



from copy import copy
from functools import singledispatch
from itertools import groupby
from typing import Callable, Dict, List, Any
import pprint
from .types import KT, VT, WILDCARD


def inspect(obj):
    pp = pprint.PrettyPrinter()
    pp.pprint(obj)


def list_flatten(t: List[List[Any]]) -> List[Any]:
    return [item for sublist in t for item in sublist]


def list_groupby(iterable: List[Any], projection) -> List[List[Any]]:
    return [list(it) for k, it in groupby(sorted(iterable, key=projection), projection)]


def list_uniq(a: List[Any]) -> List[Any]:
    return list(dict.fromkeys(a))


def dict_merge(
    a: Dict[KT, VT], b: Dict[KT, VT], f: Callable[[VT, VT], VT]
) -> Dict[KT, VT]:
    if b is None:
        return a

    merged = copy(a)
    for k in b.keys():
        if k in merged.keys():
            merged[k] = f(a[k], b[k])
        else:
            merged[k] = b[k]
    return merged


def dict_merge_with_wildcard(
    a: Dict[KT, VT], b: Dict[KT, VT], f: Callable[[VT, VT], VT]
) -> Dict[KT, VT]:
    if b is None:
        return a

    merged = copy(b)
    wb = b.get(WILDCARD, None)
    for k in a.keys():
        if k in b.keys():
            merged[k] = f(a[k], b[k])
        else:
            merged[k] = f(a[k], wb)

    wa = a.get(WILDCARD, None)
    if wa is not None:
        for k in b.keys():
            if k not in a.keys():
                merged[k] = f(wa, b[k])
    return merged


@singledispatch
def config_merger(_, b):
    return b


@config_merger.register
def _(a: dict, b: dict):
    return dict_merge(a, b, config_merger)


@config_merger.register
def _(a: list, b: list):
    return a + b



from collections import OrderedDict
from copy import copy, deepcopy
from typing import Optional

from .source import ConfigSource
from .types import ConfigDict, ConfigKey, ConfigValue
from .utils import dict_merge, config_merger


class ConfigSnippet(OrderedDict):
    def __init__(self, config: ConfigDict, source: ConfigSource):
        super().__init__(config)
        self.source = source

    def __add__(self, other: "ConfigSnippet") -> "ConfigSnippet":
        if other is None:
            return copy(self)

        return ConfigSnippet(
            config=dict_merge(self, other, config_merger),
            source=self.source + other.source,
        )

    def __copy__(self) -> "ConfigSnippet":
        return ConfigSnippet(
            config={k: copy(v) for k, v in self.items()}, source=self.source
        )

    def __deepcopy__(self, memodict={}) -> "ConfigSnippet":
        return ConfigSnippet(
            config={k: deepcopy(v, memodict) for k, v in self.items()},
            source=self.source,
        )

    def find(self, key: ConfigKey) -> Optional[ConfigValue]:
        v = self.get(key, None)
        if v is not None:
            return v
        return self.source.find(key)
